#!/usr/bin/python3

import os
import time
import random
import sys
import inquirer
##
from .snarf import Snarf
from .lib.driver import Driver
from .classes.profile import Profile
from .util.colorize import colorize
from .util.settings import Settings

# from .util.args import parser
# print(parser)
# parser.add_parser('menu', help='> access the cli menu')

####################
##### CLI Menu #####
####################

class Menu:

    ASCII = "\n     ________         .__          _________                     _____ \n \
    \\_____  \\   ____ |  | ___.__./   _____/ ____ _____ ________/ ____\\\n \
     /   |   \\ /    \\|  |<   |  |\\_____  \\ /    \\\\__  \\\\_   _ \\   __\\ \n \
    /    |    \\   |  \\  |_\\___  |/        \\   |  \\/ __ \\ |  |\\/| |   \n \
    \\_______  /___|  /____/ ____/_______  /___|  (____  \\\\__|  |_|   \n \
            \\/     \\/     \\/            \\/     \\/     \\/              \n"

    def __init__(self):
        pass

    def ask_action():
        """
        Ask action to take

        Returns
        -------
        str
            The action selected

        """

        options = ["back"]
        options.extend(Settings.get_actions())
        questions = [
            inquirer.List('action',
                message= "Please select an action:",
                choices= [str(option).title() for option in options],
            )
        ]
        answers = inquirer.prompt(questions)
        return answers['action'].lower()

    def action_menu():
        """
        Prompt the action menu. Cycles back to main menu


        """

        action = Menu.ask_action()
        if (action == 'back'): return Menu.main()
        elif (action == 'discount'): Snarf.discount()
        elif (action == 'message'): Snarf.message()
        elif (action == 'post'): Snarf.post()
        elif (action == 'profile'): Profile.menu()
        elif (action == 'promotion'): Snarf.promotion()
        else: Settings.print("Missing Action: {}".format(colorize(action,"red")))
        Menu.main()
        
    def header():
        """
        Show the header text


        """

        if not Settings.is_debug(): os.system('clear')
        print(colorize(Menu.ASCII, 'header'))
        print(colorize('version {}\n'.format(Settings.get_version()), 'green'))
        Menu.user_header()
        Menu.settings_header()

    def settings_header():
        """
        Show the settings header text


        """

        Settings.header()

    def user_header():
        """
        Show the user header text


        """

        Settings.print("User:")
        if Settings.get_username_onlyfans() != "":
            Settings.print(" - Email = {}".format(Settings.get_username_onlyfans()))
        Settings.print(" - Username = {}".format(Settings.get_username()))
        pass_ = ""
        if str(Settings.get_password()) != "":
            pass_ = "******"
        Settings.print(" - Password = {}".format(pass_))
        if str(Settings.get_username_twitter()) != "":
            Settings.print(" - Twitter = {}".format(Settings.get_username_twitter()))
            pass_ = ""
            if str(Settings.get_password_twitter()) != "":
                pass_ = "******"
            Settings.print(" - Password = {}".format(pass_))
        Settings.print('\r')

    def menu():
        """
        Prompt the basic menu selection

        Returns
        -------
        str
            The menu option selected

        """

        questions = [
            inquirer.List('menu',
                message= "Please select an option:",
                choices= ['Action', 'Settings', 'Exit']
            )
        ]
        answers = inquirer.prompt(questions)

        # menu_prompt = {
        #     'type': 'list',
        #     'name': 'menu',
        #     'message': 'Please select an option:',
        #     'choices': ['Action', 'Profile', 'Settings', 'Exit']
        # }
        # answers = prompt(menu_prompt)

        return answers['menu']

    def main_menu():
        """
        Show the main menu


        """

        action = Menu.menu()
        if (action == 'Action'): Menu.action_menu()
        elif (action == 'Settings'): Settings.menu()
        else: exit()
        Menu.main()

    def main():
        """
        Primary script entry


        """

        time.sleep(1)
        try:
            Menu.header()
            Menu.settings_header()
            Menu.main_menu()
        except Exception as e:
            Settings.dev_print(e)

#################################################################################################

def exit_handler():
    """Exit cleanly"""

    try:
        Driver.exit_all()
    except Exception as e:
        print(e)

import atexit
atexit.register(exit_handler)

######################################################

def main():
    try:
        Menu.main()
    except Exception as e:
        Settings.dev_print(e)
        Settings.print("shnarf??")
    finally:
        Settings.print("shnarrf!")
        exit_handler()

if __name__ == "__main__":
    main()


import os
import logging
from pathlib import Path
from . import defaults as DEFAULT
from .config import config

loglevel = logging.INFO
if config["debug"]: loglevel = logging.DEBUG
if int(config["verbose"]) >= 2: loglevel = logging.DEBUG

logPath = DEFAULT.LOG_PATH
if os.environ.get('ENV') == "test":
    logPath = os.path.join(os.getcwd(), "log", "snarf.log")

Path(os.path.dirname(logPath)).mkdir(parents=True, exist_ok=True)

# set up logging to file - see previous section for more details
logging.basicConfig(level=loglevel,
                    format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
                    datefmt='%m-%d %H:%M',
                    filename=logPath,
                    filemode='w')

# https://stackoverflow.com/questions/384076/how-can-i-color-python-logging-output
class CustomFormatter(logging.Formatter):
    """Logging Formatter to add colors and count warning / errors"""

    teal = ""
    grey = "\x1b[38;21m"
    yellow = "\x1b[33;21m"
    red = "\x1b[31;21m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    # the filename & line isn't helpful when i'm redirecting through Settings.maybe_print & dev_print
    # format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s (%(filename)s:%(lineno)d)"
    format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"

    FORMATS = {
        logging.DEBUG: grey + format + reset,
        logging.INFO: grey + format + reset,
        logging.WARNING: yellow + format + reset,
        logging.ERROR: red + format + reset,
        logging.CRITICAL: bold_red + format + reset
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)

# define a Handler which writes INFO messages or higher to the sys.stderr
console = logging.StreamHandler()
# tell the handler to use this format
console.setFormatter(CustomFormatter())
# add the handler to the root logger
logging.getLogger('').addHandler(console)


import pkg_resources
import time
import os, json, sys
from datetime import datetime
from pathlib import Path
##
from .colorize import colorize
from .config import config
from . import defaults as DEFAULT
from .validators import valid_schedule, valid_time
from .logger import logging
log = logging.getLogger('onlysnarf')

class Settings:
    
    LAST_UPDATED_KEY = None
    CATEGORY = None
    FILES = None
    PERFORMER_CATEGORY = None

    #####################
    ##### Functions #####
    #####################

    def debug_delay_check():
        if str(Settings.is_debug()) == "True" and str(Settings.is_debug_delay()) == "True":
            Settings.dev_print("napping...")
            time.sleep(10)

    ##
    # Print
    ##

    def print(text):
        if int(config["verbose"]) >= 0:
            log.info(text)

    def print_same_line(text):
        sys.stdout.write('\r')
        sys.stdout.flush()
        sys.stdout.write(text)
        sys.stdout.flush()

    def maybe_print(text):
        if int(config["verbose"]) >= 1 or str(os.environ.get("ENV")) == "test":
            log.debug(text)

    def dev_print(text):
        if int(config["verbose"]) >= 2 or str(os.environ.get("ENV")) == "test":
            log.debug(text)

    def err_print(error):
        log.error(error)

    def warn_print(error):
        log.warning(error)

    def format_date(date):
        if isinstance(date, str):
            return datetime.strptime(date, DEFAULT.DATE_FORMAT).strftime(DEFAULT.DATE_FORMAT)
        else:
            return date.strftime(DEFAULT.DATE_FORMAT)

    def format_time(time):
        if isinstance(time, str):
            return datetime.strptime(time, DEFAULT.TIME_FORMAT).strftime(DEFAULT.TIME_FORMAT)
        else:
            return time.strftime(DEFAULT.TIME_FORMAT)

    def header():
        Settings.print("### SETTINGS ###")
        Settings.print("...")
        Settings.print("...")
        Settings.print("...")

    def menu():
        Settings.print("### SETTINGS MENU ###")
        Settings.print("...")
        Settings.print("...")
        Settings.print("...")

    ##
    # Getters
    ##

    def get_action():
        return config["action"]

    def get_actions():
        return DEFAULT.ACTIONS

    def get_amount():
        return config["amount"]

    def get_base_directory():
        return DEFAULT.ROOT_PATH

    def get_browser_type():
        return config["browser"]

    def get_months():
        return config["months"]

    def get_category():
        cat = config["category"]
        if str(cat) == "image": cat = "images"
        if str(cat) == "gallery": cat = "galleries"
        if str(cat) == "video": cat = "videos"
        if str(cat) == "performer": cat = "performers"
        return cat or None

    def get_categories():
        cats = []
        cats.extend(list(DEFAULT.CATEGORIES))
        cats.extend(list(config["categories"]))
        return cats

    def get_cookies_path():
        # return os.path.join(Settings.get_base_directory(), Settings.get_username(), "cookies.pkl")
        return os.path.join(Settings.get_base_directory(), "cookies.pkl")

    def get_price():
        try: return config["price"]
        except Exception as e: pass
        return 0

    def get_price_minimum():
        return DEFAULT.PRICE_MINIMUM

    def get_price_maximum():
        return DEFAULT.PRICE_MAXIMUM

    def get_date():
        try:
            config["date"] = Settings.format_date(config["date"])
            if str(config["date"]) == DEFAULT.DATE and str(config["schedule"]) != DEFAULT.SCHEDULE and str(config["schedule"] != "None"):
                if isinstance(config["schedule"], str):
                    config["date"] = datetime.strptime(config["schedule"], DEFAULT.SCHEDULE_FORMAT).date().strftime(DEFAULT.DATE_FORMAT)
                else:
                    config["date"] = config["schedule"].date().strftime(DEFAULT.DATE_FORMAT)
                config["date"] = datetime.strptime(str(config["date"]), DEFAULT.DATE_FORMAT)
            else:
                config["date"] = datetime.strptime(str(config["date"]), DEFAULT.DATE_FORMAT)
            config["date"] = config["date"].strftime(DEFAULT.DATE_FORMAT)    
            Settings.maybe_print("date (settings): {}".format(config["date"]))
            return config["date"]
        except Exception as e:
            pass
        return datetime.strptime(DEFAULT.DATE, DEFAULT.DATE_FORMAT)

    def get_default_greeting():
        return DEFAULT.GREETING or ""

    def get_default_refresher():
        return DEFAULT.REFRESHER or ""
        
    def get_discount_max_amount():
        return DEFAULT.DISCOUNT_MAX_AMOUNT or 0
        
    def get_discount_min_amount():
        return DEFAULT.DISCOUNT_MIN_AMOUNT or 0
        
    def get_discount_max_months():
        return DEFAULT.DISCOUNT_MAX_MONTHS or 0
        
    def get_discount_min_months():
        return DEFAULT.DISCOUNT_MIN_MONTHS or 0

    def get_download_max():
        return config["download_limit"] or DEFAULT.IMAGE_LIMIT

    def get_duration():
        try: return config["duration"]
        except Exception as e: pass
        return None

    def get_promo_duration():
        try: return config["duration_promo"]
        except Exception as e: pass
        return None
        
    def get_duration_allowed():
        return DEFAULT.DURATION_ALLOWED or []
        
    def get_duration_promo_allowed():
        return DEFAULT.PROMOTION_DURATION_ALLOWED or []

    def get_expiration():
        try: return config["expiration"]
        except Exception as e: pass
        return DEFAULT.EXPIRATION_NONE

    def get_promo_expiration():
        return config["promotion_expiration"]

    def get_input():
        # fix pytest bug from 4.4.9
        files = []
        for file_path in config["input"]:
            if ".py" not in str(file_path):
                files.append[file_path]
        return set(files)

    def get_input_as_files():
        if Settings.FILES: return Settings.FILES
        from ..classes.file import File
        files = []
        try:
            _input = config["input"]
            if isinstance(_input, list):
                for file_path in _input:
                    file = File()
                    setattr(file, "path", file_path)
                    files.append(file)
            else:
                file = File()
                setattr(file, "path", _input)
                files.append(file)
        except Exception as e:
            pass
        Settings.FILES = files
        return files
        
    def get_logs_path(process):
        if process == "firefox":
            path_ = os.path.join(Settings.get_base_directory(), "log")
            Path(path_).mkdir(parents=True, exist_ok=True)
            return os.path.join(path_, "geckodriver.log")
        elif process == "google":
            path_ = os.path.join(Settings.get_base_directory(), "log")
            Path(path_).mkdir(parents=True, exist_ok=True)
            return os.path.join(path_, "chromedriver.log")
        return ""

    def get_message_choices():
        return DEFAULT.MESSAGE_CHOICES

    def get_root_path():
        return DEFAULT.ROOT_PATH

    def get_sort_method():
        return config["sort"] or "random"

    def get_phone_number():
        try:
            return config["phone"]
        except Exception as e:
            Settings.err_print("missing phone number!")
        return None

    def get_performers():
        try:
            performers = config["performers"] or []
            performers = [n.strip() for n in performers]
            return performers
        except Exception as e: pass
        return []

    def get_profile_path():
        try: return config["path_profile"]
        except Exception as e: pass
        return DEFAULT.PROFILE_PATH

    def get_recent_user_count():
        return config["recent_users_count"] or 0
    
    def get_promotion_limit():
        return config["promotion_limit"] or None

    def get_promotion_method():
        return config["promotion_method"] or None

    def get_password():
        try: return Settings.get_user_config(Settings.get_username())["onlyfans_password"]
        except Exception as e: Settings.err_print(e)
        return ""

    def get_password_google():
        try: return Settings.get_user_config(Settings.get_username())["google_password"]
        except Exception as e: Settings.err_print(e)
        return ""

    def get_password_twitter():
        try: return Settings.get_user_config(Settings.get_username())["twitter_password"]
        except Exception as e: Settings.err_print(e)
        return ""

    def get_download_path():
        try: return config["path_download"]
        except Exception as e: pass
        return DEFAULT.DOWNLOAD_PATH

    def get_users_path():
        try: return config["path_users"]
        except Exception as e: pass
        return DEFAULT.USERS_PATH

    def get_config_path():
        try: return config["path_config"]   
        except Exception as e: pass
        return DEFAULT.CONFIG_PATH

    def get_local_path():
        localPath = os.path.join(Settings.get_root_path(), Settings.get_username())
        from pathlib import Path
        Path(localPath).mkdir(parents=True, exist_ok=True)
        for cat in Settings.get_categories():
            Path(os.path.join(localPath, cat)).mkdir(parents=True, exist_ok=True)
        return localPath

    def get_reconnect_id():
        return config["session_id"] or ""

    def get_reconnect_url():
        return config["session_url"] or ""

    def get_remote_host():
        return config["remote_host"] or DEFAULT.REMOTE_HOST

    def get_remote_port():
        return config["remote_port"] or DEFAULT.REMOTE_PORT

    def get_remote_path():
        return config["remote_path"] or DEFAULT.REMOTE_PATH

    def get_remote_username():
        return config["remote_username"] or ""

    def get_remote_password():
        return config["remote_password"] or ""

    def get_profile_method():
        return config["profile_method"] or None

    def get_schedule():
        schedule = ""
        try:
            schedule = config["schedule"]
            if str(schedule) == "None": schedule = DEFAULT.SCHEDULE
            if str(schedule) == DEFAULT.SCHEDULE:
                schedule = datetime.strptime(schedule, DEFAULT.SCHEDULE_FORMAT).strftime(DEFAULT.SCHEDULE_FORMAT)
            elif not isinstance(schedule, str):
                schedule = schedule.strftime(DEFAULT.SCHEDULE_FORMAT)
            Settings.maybe_print("schedule (settings): {}".format(schedule))
            return schedule
        except Exception as e:
            pass
        return datetime.strptime("{} {}".format(Settings.get_date(), Settings.get_time()), DEFAULT.SCHEDULE_FORMAT).strftime(DEFAULT.SCHEDULE_FORMAT)

    def get_keywords():
        keywords = []
        try:
            keywords = config["keywords"]
            keywords = [n.strip() for n in keywords]
        except Exception as e:
            pass
            # Settings.err_print(e)
        return keywords

    def get_text():
        return config["text"] or ""

    def get_time():
        try:
            config["time"] = Settings.format_time(config["time"])        
            if (str(config["time"]) == DEFAULT.TIME or str(config["time"]) == DEFAULT.TIME_NONE) and str(config["schedule"]) != DEFAULT.SCHEDULE and str(config["schedule"]) != "None":
                Settings.dev_print("time from schedule")
                date = datetime.strptime(str(config["schedule"]), DEFAULT.SCHEDULE_FORMAT)
                config["time"] = datetime.strptime(str(date.time().strftime(DEFAULT.TIME_FORMAT)), DEFAULT.TIME_FORMAT)
            else:
                Settings.dev_print("time from config")
                config["time"] = datetime.strptime(str(config["time"]), DEFAULT.TIME_FORMAT)
            config["time"] = config["time"].strftime(DEFAULT.TIME_FORMAT)
            Settings.maybe_print("time (settings): {}".format(config["time"]))
            return config["time"]
        except Exception as e:
            pass
        return datetime.strptime(DEFAULT.TIME, DEFAULT.TIME_FORMAT).strftime(DEFAULT.TIME_FORMAT)

    def get_title():
        return config["title"] or ""
        
    def get_skipped_users():
        return config["skipped_users"] or []
        
    def get_questions():
        try:
            return config["questions"]
        except Exception as e:
            pass
        return []
        
    def get_upload_max():
        try:
            return config["upload_max"]
        except Exception as e:
            pass
        return DEFAULT.IMAGE_LIMIT
        
    # def get_upload_max_messages():
        # return config["upload_max_messages"] or UPLOAD_MAX_MESSAGES

    def get_login_method():
        try: return config["login"]
        except Exception as e: pass
        return "auto"
        
    def get_upload_max_duration():
        return config["upload_max_duration"] or DEFAULT.UPLOAD_MAX_DURATION # 1 hour

    # comma separated string of usernames
    def get_users():
        from ..classes.user import User
        if str(config["user"]) != "None":
            if str(config["user"]) == "all":
                config["users"].extend([user.username for user in User.get_all_users()])
            elif str(config["user"]) == "recent":
                config["users"].extend([user.username for user in User.get_recent_users()])
            elif str(config["user"]) == "favorite":
                config["users"].extend([user.username for user in User.get_favorite_users()])
            elif str(config["user"]) == "random":
                config["users"] = [User.get_random_user().username]
            else: config["users"].append(config["user"])
        users = []
        for user in [n.strip() for n in config["users"]]:
            if isinstance(user, User): pass
            elif isinstance(user, str): user = User({"username":user})
            # BUG (potential): might bug out if the username is for whatever reason all numbers
            elif isinstance(user, int): user = User({"id":user})
            users.append(user)
        return users

    def get_user():
        return Settings.get_users()[0]

    def get_user_configs():
        # load configs from .onlysnarf or baseDir
        pass

    def get_user_config(username="default"):
        import configparser
        config_file = configparser.ConfigParser()
        # strip email
        if "@" in username: username = username[0 : username.index("@")]
        Settings.dev_print("retrieving user config: {}".format(username))
        Settings.dev_print(os.path.join(Settings.get_base_directory(), "conf", "users", username+".conf"))
        config_file.read(os.path.join(Settings.get_base_directory(), "conf", "users", username+".conf"))
        userConfig = {}
        for section in config_file.sections():
            # Settings.dev_print(section)
            for key in config_file[section]:
                # Settings.dev_print(section, key, config_file[section][key].strip("\""))
                userConfig[section.lower()+"_"+key.lower()] = config_file[section][key].strip("\"")
        # Settings.dev_print(userConfig)
        return userConfig

    def get_username():
        return config["username"] or "default"

    def get_username_onlyfans():
        try:
            return Settings.get_user_config(Settings.get_username())["onlyfans_username"]
        except Exception as e:
            Settings.err_print(e)
        return ""

    def get_username_google():
        try:
            return Settings.get_user_config(Settings.get_username())["google_username"]
        except Exception as e:
            Settings.err_print(e)
        return ""            

    def get_username_twitter():
        try:
            return Settings.get_user_config(Settings.get_username())["twitter_username"]
        except Exception as e:
            Settings.err_print(e)
        return ""

    def get_remote_browser_host():
        return config["remote_browser_host"]

    def get_remote_browser_port():
        return config["remote_browser_port"]

    ## TODO
    # add arg -profile
    # add method for reading config profiles from conf/users

    def get_profile():
        pass

    def select_profile():
        pass

    # def get_users_favorite():
    #     return config["users_favorite"] or []
        
    def get_verbosity():
        return config["verbose"] or 0

    def get_version():
        return pkg_resources.get_distribution("onlysnarf").version

    def get_user_num():
        return config["users_read"] or DEFAULT.USER_LIMIT

    # Bools

    def is_confirm():
        return config["confirm"]

    def is_cookies():
        return config["cookies"]

    def is_delete():
        return config["delete"]

    def is_delete_empty():
        return config["delete_empty"]

    def is_debug(process=None):
        if process == "firefox": return config["debug_firefox"]
        elif process == "chrome": return config["debug_chrome"]
        elif process == "selenium": return config["debug_selenium"]
        elif process == "cookies": return config["debug_cookies"]
        # elif process == "tests": return 
        return config["debug"]

    def is_debug_delay():
        return config["debug_delay"]

    def is_force_upload():
        return config["force_upload"]

    def is_keep():
        return config["keep"]

    def is_prefer_local():
        return config["prefer_local"]
        
    def is_prefer_local_following():
        return config["prefer_local_following"]

    def is_save_users():
        return config["save_users"]
        
    def is_reduce():
        return config["reduce"]
    
    def is_show_window():
        return config["show"]
        
    def is_tweeting():
        return config["tweeting"]
        
    def is_backup():
        return config["backup"]
        
    def is_skip_download():
        return config["skip_download"]
        
    def is_skip_upload():
        return config["skip_upload"]

    ##
    # Setters
    ##

    def set_bycategory(cat):
        config["bycategory"] = cat

    def set_category(cat):
        config["category"] = cat

    def set_cookies(value):
        config["cookies"] = value

    def set_confirm(value):
        config["confirm"] = value

    def set_email(email):
        config["email"] = str(email)

    def set_debug(newValue):
        if str(newValue) == "tests":
            pass
            # config["confirm"] = False
        else:
            config["debug"] = newValue

    def set_username(username):
        config["username"] = str(username)

    def set_username_google(username):
        config["username_google"] = str(username)

    def set_username_twitter(username):
        config["username_twitter"] = str(username)

    def set_password(password):
        config["password"] = str(password)

    def set_password_google(password):
        config["password_google"] = str(password)

    def set_password_twitter(password):
        config["password_twitter"] = str(password)

    def set_prefer_local(buul):
        config["prefer_local"] = buul
    
    def set_prefer_local_following(buul):
        config["prefer_local_following"] = buul

###########################################################################



#     def update_value(self, variable, newValue):
#         variable = str(variable).upper().replace(" ","_")
#         try:
#             # print("Updating: {} = {}".format(variable, newValue))
#             setattr(self, variable, newValue)
#             # print("Updated: {} = {}".format(variable, getattr(self, variable)))
#         except Exception as e:
#             maybePrint(e)

# # move this behavior to user
#     def update_profile_value(self, variable, newValue):
#         variable = str(variable).upper().replace(" ","_")
#         try:
#             # print("Updating: {} = {}".format(variable, newValue))
#             Settings.PROFILE.setattr(self, variable, newValue)
#             # print("Updated: {} = {}".format(variable, getattr(self, variable)))
#         except Exception as e:
#             maybePrint(e)














































#######################################################################################

def delayForThirty():
    Settings.maybe_print("30...")
    time.sleep(10)
    Settings.maybe_print("20...")
    time.sleep(10)
    Settings.maybe_print("10...")
    time.sleep(7)
    Settings.maybe_print("3...")
    time.sleep(1)
    Settings.maybe_print("2...")
    time.sleep(1)
    Settings.maybe_print("1...")
    time.sleep(1)


#!/usr/bin/python3

from .lib.driver import Driver
from .util.settings import Settings
from .classes.discount import Discount
from .classes.message import Message, Post
from .classes.profile import Profile
from .classes.promotion import Promotion
from .classes.user import User

#################
##### Snarf #####
#################

class Snarf:

    """
    OnlySnarf main class and runtime parser.

    All methods are static and handle the basic runtime operations, 
     importing variables from settings & args.

    """

    def __init__(self):
        """Snarf object"""

        pass

    @staticmethod
    def close():
        Driver.exit_all()
        Settings.print("*snarf waves goodbye*")

    @staticmethod
    def discount():

        """
        Applies a discount to users as provided from args / prompts.


        """

        try:
            successful = []
            for user in Settings.get_users():
                Settings.print("> Discounting fan: {}".format(user.username))
                discount = Discount(user.username)
                successful.append(discount.apply())
            return all(successful)
        except Exception as e: Settings.dev_print(e)
        return False

    @staticmethod
    def message():

        """
        Sends the configured message from args / prompts.

        
        """

        try:
            successful = []
            for user in Settings.get_users():
                Settings.print("> Messaging fan: {}".format(user.username))
                message = Message(user.username)
                successful.append(message.send(user.username, user_id=user.id))
            return all(successful)
        except Exception as e: Settings.dev_print(e)
        return False
                
    @staticmethod
    def post():

        """
        Posts the configured text from args / prompts.

        
        """

        post = Post()
        try: return post.send()
        except Exception as e: Settings.dev_print(e)
        return False

    @staticmethod
    def profile():

        """
        Runs the profile method specified at runtime.

        backup - downloads all content and saves settings

        syncFrom - reads all profile settings and saves locally

        syncTo - updates profile settings with provided profile

        Extended description of function.

        """

        profile = Profile()
        try: 
            # get profile method
            method = Settings.get_profile_method()
            if method == "backup": return Profile.backup_content()
            elif method == "syncfrom": return Profile.sync_from_profile()
            elif method == "syncto": return Profile.sync_to_profile()
            else: Settings.err_print("Missing Profile Method")
        except Exception as e: Settings.dev_print(e)
        return False
        
    @staticmethod
    def promotion():

        """
        Runs the promotion method specified at runtime.

        campain - creates discount campaign

        trial - creates free trial

        user - applies directly to user

        grandfather - applies discounted price to existing users and adds them all to list

        """

        try: 
            # get promotion method
            method = Settings.get_promotion_method()
            if method == "campaign": return Promotion.create_campaign()
            elif method == "trial": return Promotion.create_trial_link()
            elif method == "user": return Promotion.apply_to_user()
            elif method == "grandfather": return Promotion.grandfathered()
            else: Settings.err_print("Missing Promotion Method")
        except Exception as e: Settings.dev_print(e)
        return False

    @staticmethod
    def users():

        """
        Scan users.

        
        """

        try:
            Settings.set_prefer_local(False)
            User.get_all_users()
            return True
        except Exception as e: Settings.dev_print(e)
        return False

################################################################################################################################################

# import sys

def exit_handler():
    """Exit cleanly"""

    try:
        Driver.exit_all()
        # sys.exit(0)
    except Exception as e:
        print(e)

import atexit
atexit.register(exit_handler)

def main():
    try:
        # purge local tmp files
        from .classes.file import File
        File.remove_local()
        # get the thing, do the thing
        action = Settings.get_action()
        Settings.print("Running - {}".format(action))
        action = getattr(Snarf, action)
        action()
    except Exception as e:
        Settings.dev_print(e)
        Settings.print("shnarf??")
    finally:
        Settings.print("shnarrf!")
        exit_handler()

################################################################################################################################################

if __name__ == "__main__":
    main()


import os, shutil, random, sys
from os import walk
##
from ..lib.ffmpeg import ffmpeg
from ..util.settings import Settings
import wget

###############################################################

class File():
    """File class for manipulating files."""

    ONE_GIGABYTE = 1000000000
    ONE_MEGABYTE = 1000000
    FIFTY_MEGABYTES = 50000000
    ONE_HUNDRED_KILOBYTES = 100000
    
    MIMETYPES_IMAGES = "(mimeType contains 'image/jpeg' or mimeType contains 'image/jpg' or mimeType contains 'image/png')"
    MIMETYPES_VIDEOS = "(mimeType contains 'video/mp4' or mimeType contains 'video/quicktime' or mimeType contains 'video/x-ms-wmv' or mimeType contains 'video/x-flv')"
    MIMETYPES_ALL = "(mimeType contains 'image/jpeg' or mimeType contains 'image/jpg' or mimeType contains 'image/png' or mimeType contains 'video/mp4' or mimeType contains 'video/quicktime')"
    MIMETYPES_IMAGES_LIST = ["image/jpeg","image/jpg","image/png"]
    MIMETYPES_VIDEOS_LIST = ["video/mp4","video/quicktime","video/x-ms-wmv","video/x-flv"]
    MIMETYPES_ALL_LIST = []
    MIMETYPES_ALL_LIST.extend(MIMETYPES_IMAGES_LIST)
    MIMETYPES_ALL_LIST.extend(MIMETYPES_VIDEOS_LIST)

    def __init__(self):
        """File object represents local image/video file"""

        # the path to the file locally
        self.path = ""
        # the file extension
        self.ext = ""
        # image|video, default image
        self.type = "image"
        ##
        # file title reference
        self.title = ""
        # file size
        self.size = 0

    ######################################################################################

    def check_size(self):
        """
        Check file size.

        Returns
        -------
        bool
            Whether or not the file exists by checking size

        """

        size = self.size
        if not size and not os.path.exists(self.get_path()):
            return False
        if size: return True
        size = os.path.getsize(self.get_path())
        Settings.maybe_print("file size: {}kb - {}mb".format(size/1000, size/1000000))
        if size <= File.ONE_HUNDRED_KILOBYTES:
            Settings.warn_print("tiny file size")
        elif size <= File.ONE_MEGABYTE:
            Settings.warn_print("small file size")
        elif size > 0:
            Settings.maybe_print("normal file size")
        else:
            Settings.err_print("empty file size")
            return False
        self.size = size
        return True

    ##############################

    def download(self):
        """Download a url. An input can only be a valid path or a valid url."""

        Settings.maybe_print("downloading file...")
        filename = wget.download(self.path, out=self.get_tmp())
        Settings.print("") # resume same line after wget download
        Settings.maybe_print("downloaded: "+filename)
        self.path = filename

    def get_ext(self):
        """Get the file's extension"""

        if self.ext != "": return self.ext
        self.get_title()
        return self.ext

    def get_path(self):
        """
        Get the file's path
        
        Returns
        -------
        str
            The file path

        """

        if self.path == "":
            Settings.err_print("missing file path")
            return  ""
        return str(self.path)

    def get_title(self):
        """
        Get the file's title from it's filename
        
        Returns
        -------
        str
            The file's title or filename without extension

        """

        if self.title != "": return self.title
        path = self.get_path()
        if str(path) == "": 
            Settings.err_print("missing file title!")
            return ""
        title, ext = os.path.splitext(path)
        self.ext = ext.replace(".","")
        self.title = "{}{}".format(os.path.basename(title), ext)
        return self.title

    @staticmethod
    def get_tmp():
        """Creates / gets the default temporary download directory"""

        download_path = Settings.get_download_path()
        if not os.path.exists(download_path):
            os.mkdir(download_path)
        return download_path

    def get_type(self):
        """
        Gets the file's type as an inner class of either Image or Video
        
        Returns
        -------
        Image|Video
            The file's type as an image or video class

        """

        if self.type: return self.type
        if str(self.get_ext()) in str(File.MIMETYPES_VIDEOS_LIST):
            self.type = Video()
        elif str(self.get_ext()) in str(File.MIMETYPES_IMAGES_LIST):
            self.type = Image()
        else: Settings.warn_print("unable to parse file type")
        return self.type

    def prepare(self):
        """
        Prepares the file for uploading.

        Runs the apppropriate file type method and downloads the file locally if necessary.

        Returns
        -------
        bool
            Whether or not the file is prepared

        """

        Settings.maybe_print("preparing file: {}".format(self.get_title()))
        # self.get_type().prepare()
        if not self.check_size():
            self.download()
        return self.check_size()

    @staticmethod
    def get_files_by_folder(path):
        """
        Get local files from the local folder path.

        Parameters
        ----------
        path : str
            Path to folder to get files of

        Returns
        -------
        list
            The files at the path

        """

        f = []
        for (dirpath, dirnames, filenames) in walk(path):
            f.extend(filenames)
            break
        return f

    def get_random_file(self):
        """Get random file from all files"""

        return random.choice(self.get_files())

    @staticmethod
    def get_images_of_folder(folder):
        """
        Get images of folder.

        Parameters
        ----------
        folder : str
            The folder path to get images from

        Returns
        -------
        list
            The discovered image files

        """

        Settings.dev_print("getting images of folder: {}".format(folder.get_title()))
        if not folder: return []
        imgs = []
        files = []
        valid_images = [".jpg",".gif",".png",".tga",".jpeg"]
        for f in os.listdir(folder.get_path()):
            ext = os.path.splitext(f)[1]
            if ext.lower() not in valid_images:
                continue
            file = File()
            setattr(file, "path", os.path.join(folder.get_path(),f))
            files.append(file)
            Settings.maybe_print("image path: {}".format(os.path.join(folder.get_path(),f)))
        return files

    @staticmethod
    def get_videos_of_folder(folder):
        """
        Get videos of folder.

        Parameters
        ----------
        folder : str
            The folder path to get videos from

        Returns
        -------
        list
            The discovered video files

        """

        Settings.dev_print("getting videos of folder: {}".format(folder.get_title()))
        if not folder: return []
        videos = []
        files = []

        ## TODO: change this to mimetypes

        valid_videos = [".mp4",".mov"]
        for f in os.listdir(folder.get_path()):
            ext = os.path.splitext(f)[1]
            if ext.lower() not in valid_videos:
                continue
            file = File()
            setattr(file, "path", os.path.join(folder.get_path(),f))
            files.append(file)
            Settings.maybe_print("video path: {}".format(os.path.join(folder.get_path(),f)))
        return files

    @staticmethod
    def get_folders_of_folder(folderPath):
        """
        Get folders of folder.

        Parameters
        ----------
        folderPath : str
            The folder path to get folders from

        Returns
        -------
        list
            The discovered folders

        """

        # os.walk(directory)
        # will yield a tuple for each subdirectory. Ths first entry in the 3-tuple is a directory name, so
        # [x[0] for x in os.walk(directory)]
        # should give you all of the subdirectories, recursively.
        # Note that the second entry in the tuple is the list of child directories of the entry in the first position, so you could use this instead, but it's not likely to save you much.
        # However, you could use it just to give you the immediate child directories:
        Settings.maybe_print("local walk: {}".format(folderPath))
        folders = []
        # Settings.print(os.walk(folderPath))
        for folder in next(os.walk(folderPath))[1]:
            Settings.maybe_print("folder: {}".format(folder))
            fol = Folder()
            setattr(fol, "path", os.path.join(folderPath, folder))
            folders.append(fol)
        return folders

    @staticmethod
    def remove_local():
        """
        Delete all local files.

        """

        try:
            Settings.maybe_print('deleting local files...')
            # delete /tmp
            tmp = File.get_tmp()
            if os.path.exists(tmp):
                shutil.rmtree(tmp)
                Settings.maybe_print('local files removed!')
            else:
                Settings.maybe_print('no local files found!')
        except Exception as e:
            Settings.dev_print(e)

    # def upload(self):
    #     """
    #     Process ran by a file after it has been uploaded.

    #     Ensures the file has been backed up and then deleted locally.
        
    #     Returns
    #     -------
    #     bool
    #         Whether or not the file was properly handled after its upload

    #     """
    #     if not self.prepare():
    #         Settings.err_print("unable to upload file - {}".format(self.get_title()))
    #         return False
    #     return True

######################################################################################################################
######################################################################################################################
######################################################################################################################

class Folder(File):
    def __init__(self):
        File.__init__(self)
        self.files = None

    def check_size(self):
        """
        Check the size of the files in the folder to check if the folder exists.

        Returns
        -------
        bool
            Whether or not the folder exists

        """

        for file in self.get_files():
            exists = file.check_size()
            if not exists: return False
        return True

    # def combine(self):
    #     if len(self.files) == 0: return
    #     Settings.dev_print("combining files: {}".format(len(self.files)))
    #     Settings.dev_print("combine path: {}".format(combinedPath))
    #     combinedPath = os.path.join(File.get_tmp(), "{}-combined".format(self.title))
    #     for file in files:
    #         shutil.move(file.get_path(), combinedPath)
    #         file.path = "{}/{}".format(combinedPath, self.title)
    #     self.combined = ffmpeg.combine(combinedPath)

    ##############################

    def get_files(self):
        """
        Get files from the folder.


        Returns
        -------
        list
            The discovered files

        """

        if not self.files and self.path:
            self.files = []
            files = File.get_files_by_folder(self.get_path())
            for file in files:
                file_ = File()
                setattr(file_, "path", os.path.join(self.get_path(), file))
                self.files.append(file_)
                Settings.maybe_print("local file found: {}".format(file_.get_title()))
        if Settings.get_title():
            for file in self.files:
                if str(Settings.get_title()) == str(file.get_title()):
                    self.files = [file]
                    break
        return self.files

    def get_title(self):
        """
        Get the title of the folder.

        Returns
        -------
        str
            The folder's title

        """

        if self.title: return self.title
        path = self.get_path()
        if str(path) == "": 
            Settings.err_print("missing file title")
            return ""
        title = os.path.basename(path)
        self.title = title
        return self.title

    def prepare():
        """
        Prepare the files in the folder for handling.

        Returns
        -------
        bool
            Whether or not the folder has been prepared successfully

        """

        Settings.maybe_print("preparing folder: {}".format(self.get_title()))
        prepared = False
        for file in self.get_files():
            prepared_ = file.prepare()
            if prepared_: prepared = prepared_
        return prepared

######################################################################################################################
######################################################################################################################
######################################################################################################################

class Image(File):
    def __init__(self):
        pass

    def prepare(self):
        """
        Prepare the image.

        Returns
        -------
        bool
            Whether or not the image has been prepared

        """

        Settings.maybe_print("preparing image: {}".format(self.get_title()))
        return super().prepare()

######################################################################################################################
######################################################################################################################
######################################################################################################################

class Video(File):
    def __init__(self):
        self.screenshots = []
        self.trimmed = ""
        self.split = ""

    #seconds off front or back
    def trim(self):
        """Trim the video file."""

        path = self.get_path()
        self.trimmed = ffmpeg.trim(path) 

    # into segments (60 sec, 5 min, 10 min)
    def split(self):
        """Split the video file."""

        path = self.get_path()
        self.split = ffmpeg.split(path)

    # unnecessary, handled by onlyfans
    # unless this somehow adds like more metadata
    def watermark(self):
        pass

    # cleanup & label appropriately (digital watermarking?)
    # def get_metadata(self):
        # pass

    # frames for preview gallery
    def get_frames(self):
        """Get frames from the video as screenshots."""

        path = self.get_path()
        self.screenshots = ffmpeg.frames(path)

    def prepare(self):
        """
        Prepare the video.

        Returns
        -------
        bool
            Whether or not the video has been prepared

        """

        Settings.maybe_print("preparing video: {}".format(self.get_title()))
        self.reduce()
        self.repair()
        self.watermark()
        return super().prepare()

    def reduce(self):
        """Reduce the video file."""

        if not Settings.is_reduce(): 
            Settings.maybe_print("skipping: video reduction")
            return
        path = self.get_path()
        if (int(os.stat(str(path)).st_size) < File.FIFTY_MEGABYTES or str(Settings.is_reduce()) == "False"):
            return
        Settings.dev_print("reduce: {}".format(self.get_title()))
        self.path = ffmpeg.reduce(path)
    
    # unnecessary
    # def repair(self):
    #     """Repair the video file."""
        
    #     if not Settings.is_repair():
    #         Settings.dev_print("skipping: video repair")
    #         return
    #     path = self.get_path()
    #     if Settings.is_repair():
    #         return
    #     Settings.dev_print("repair: {}".format(self.get_title()))
    #     self.path = ffmpeg.repair(path)




import re
from datetime import datetime
from decimal import Decimal
from re import sub
##
from ..lib.driver import Driver
from .file import File, Folder
from .poll import Poll
from .user import User
from ..util.settings import Settings
from .schedule import Schedule

class Message():
    """OnlyFans message (and post) class"""

    def __init__(self, users=[]):
        """
        OnlyFans message and post object

        A post is just a message on a profile with different options made available. So all posts are messages, as all messages are messages.
            Squares and rectangles.

        """

        # universal message variables
        self.text = ""
        self.files = []
        self.performers = []
        self.price = 0 # $3 - $100
        self.keywords = []
        self.__initialized__ = False

    def init(self):
        """Initialize."""

        if self.__initialized__: return
        self.get_text()
        self.get_keywords()
        self.get_price()
        self.get_files()
        self.get_performers()
        self.__initialized__ = True

    @staticmethod
    def format_keywords(keywords):
        """
        Formats the list provided into a combined string with a # in front of each value.

        Parameters
        ----------
        keywords : list
            List of keywords as strings

        Returns
        -------
        str
            The generated keywords into a string
        """

        # ternary: a if condition else b
        return "#{}".format(" #".join(keywords)) if len(keywords) > 0 else ""

    @staticmethod
    def format_performers(performers):
        """
        Formats the list provided into a combined string with an @ in front of each value.
            A space is added before @ to close performer search modal (???).

        Parameters
        ----------
        performers : list
            List of performers usernames as strings

        Returns
        -------
        str
            The generated performers into a string
        """

        # ternary: a if condition else b
        return " @{} ".format(" @".join(performers)) if len(performers) > 0 else ""


    def format_text(self):
        """Formats self.text with the provided keywords and performers

        
        Returns
        -------
        str
            The generated text into a string. Example:
            "This is the text. @name, @name, and @name #keyword0 #keyword1"

        """

        return "{}{}{}".format(self.get_text(), Message.format_performers(self.get_performers()), Message.format_keywords(self.get_keywords())).strip()

    @staticmethod
    def is_tip(text):
        """
        Checks if the text contains a tip amount.

        Parameters
        ----------
        text : str
            The text to parse

        Returns
        -------
        bool
            Whether the text contains a tip or not
        int
            The tip amount contained, default 0

        """

        if re.search(r'I sent you a \$[0-9]*\.00 tip ♥', text):
            amount = re.match(r'I sent you a \$([0-9]*)\.00 tip ♥', text).group(1)
            Settings.maybe_print("message contains (tip): {}".format(amount))
            return True, int(amount)
        elif re.search(r"I\'ve contributed \$[0-9]*\.00 to your Campaign", text):
            amount = re.match(r'I\'ve contributed \$([0-9]*)\.00 to your Campaign', text).group(1)
            Settings.maybe_print("message contains (campaign): {}".format(amount))
            return True, int(amount)
        return False, 0

    def get_files(self):
        """
        Gets files from args specified source or prompts as necessary.

        Uses appropriate file select method as specified by runtime args:
        - remote (server, ipfs)
        - local

        Parameters
        ----------
        again : bool
            Whether or not it is the script user's first time around.

        Returns
        -------
        list
            Files in a list

        """

        if len(self.files) > 0:
            files_ = []
            for file in self.files[:int(Settings.get_upload_max())]:
                if not isinstance(file, File):
                    file_ = File()
                    setattr(file_, "path", file)
                    files_.append(file_)
                else:
                    files_.append(file)
            return files_
            # return self.files[:int(Settings.get_upload_max())]
        files = Settings.get_input_as_files()
        if len(files) > 0:
            Settings.dev_print("fetched input files for upload")
            self.files = files[:int(Settings.get_upload_max())] # reduce by max
            # self.files = files
            # return files
        return self.files
        # files = Folder.get_files()
        # if files is empty this all basically just skips to the end and returns blank 
        # filed = []
        # for file in files:
            # turn all folders into their files
            # if isinstance(file, Folder): filed.extend(file.get_files())
            # else:
                # filed.append(file)
                # TODO
                # this goes elsewhere
                # flag that the files include a performer
                # if hasattr(file, "performer"):
                    # self.performers.append(getattr("performer", file))

    def get_message(self):
        """
        Gets the message as a serialized JSON object.


        Returns
        -------
        Object
            The message as an object.

        """

        return dict({
            "text": self.format_text(),
            "files": self.get_files(),
            "performers": self.get_performers(),
            "price": self.get_price(),
            "keywords": self.get_keywords()
        })

    def get_performers(self):
        """
        Gets the performers for the text.

        Returns
        -------
        list
            The performers

        """

        if len(self.performers) > 0: return self.performers
        self.performers = Settings.get_performers()
        return self.performers

    def get_price(self):
        """
        Gets the price value if not none else sets it from args or prompts.


        Returns
        -------
        int
            The price

        """

        if self.price: return self.price
        price = Settings.get_price()
        if str(price) == "0": return 0
        priceMin = Settings.get_price_minimum()
        priceMax = Settings.get_price_maximum()
        if str(price) == "max": price = priceMax
        elif str(price) == "min": price = priceMin
        elif Decimal(sub(r'[^\d.]', '', str(price))) < Decimal(priceMin):
            Settings.warn_print("price too low: {} < {}".format(price, priceMin))
            Settings.maybe_print("adjusting price to minimum...")
            price = priceMin
        elif Decimal(sub(r'[^\d.]', '', str(price))) > Decimal(priceMax):
            Settings.warn_print("price too high: {} < {}".format(price, priceMax))
            Settings.maybe_print("adjusting price to maximum...")
            price = priceMax    
        self.price = price
        return self.price

    def get_keywords(self):
        """
        Gets the keywords for the text.

        Returns
        -------
        list
            The keywords

        """

        if len(self.keywords) > 0: return self.keywords
        self.keywords = Settings.get_keywords()
        return self.keywords

    def get_text(self, again=True):
        """
        Gets the text value if not none else sets it from args or prompts.


        Parameters
        ----------
        again : bool
            Whether or not it is the script user's first time around.

        Returns
        -------
        str
            The text to enter.

        """

        if self.text != "": return self.text
        # retrieve from args and return if exists
        text = Settings.get_text()
        if text != "": 
            self.text = text
            return text
        text = self.get_text_from_filename()
        if text != "":
            self.text = text
            return text
        self.text = text
        return self.text

    def get_text_from_filename(self):
        """Gets text from this object's file's title"""

        if not self.get_files(): return ""
        text = self.files[0].get_title()
        # if "_" in str(self.text):
        if re.match("[0-9]_[0-9]", text) is not None:
            texttext = self.files[0].get_parent()["title"]
        else:
            try: 
                int(text)
                # is a simple int
                if int(text) > 20:
                    text = self.files[0].get_parent()["title"]
            except Exception as e:
                # not a simple int
                # do nothing cause probably set already
                pass
        text = text.replace("_", " ")
        # redo keyword parsing (unsure if necessary call)
        text = self.update_keywords(text)
        return text

    def send(self, username, user_id=None):
        """
        Sends a message.


        Returns
        -------
        bool
            Whether or not sending the message was successful.

        """

        self.init()
        return User.message_user(self.get_message(), username, user_id=user_id)            

class Post(Message):
    """OnlyFans message (and post) class"""

    def __init__(self):
        """
        OnlyFans post object

        A post is just a message on a profile with different options made available. So all posts are messages, as all messages are messages.
            Squares and rectangles.

        """

        super().__init__(self)
        self.expiration = 0
        self.poll = None
        self.schedule = None

    # def __str__(self):
    #     return "fooPost"

    def init(self):
        """Initialize."""

        super().init()
        self.__initialized__ = False
        self.get_poll()
        self.get_schedule()
        self.get_expiration()
        self.__initialized__ = True

    def get_expiration(self, again=True):
        """
        Gets the expiration value if not none else sets it from args or prompts.
        
        Parameters
        ----------
        again : bool
            Whether or not it is the script user's first time around.


        Returns
        -------
        int
            The expiration as an int.

        """

        if self.expiration: return self.expiration
        # retrieve from args and return if exists
        expiration = Settings.get_expiration() or 0
        if expiration: 
            self.expiration = expiration
            return expiration
        self.expiration = expiration
        return self.expiration

    def get_poll(self, again=True):
        """
        Gets the poll value if not none else sets it from args or prompts.
        
        Parameters
        ----------
        again : bool
            Whether or not it is the script user's first time around.


        Returns
        -------
        Poll
            Poll object with proper values

        """

        # check if poll is ready
        if self.poll: return self.poll
        self.poll = Poll()
        return self.poll

    def get_post(self):
        """
        Gets the message as a serialized JSON object.


        Returns
        -------
        Object
            The message as an object.

        """

        return dict({
            "text": self.format_text(),
            "files": self.get_files(),
            "performers": self.get_performers(),
            "price": self.get_price(),
            "expiration": self.get_expiration(),
            "schedule": self.get_schedule(),
            "poll": self.get_poll(),
            "keywords": self.get_keywords()
        })

    def get_schedule(self):
        """
        Gets the schedule value if not none else sets it from args or prompts.

        Returns
        -------
        Schedule
            Schedule object with proper values.

        """

        if self.schedule: return self.schedule
        self.schedule = Schedule()
        return self.schedule

    def send(self):
        """
        Sends a post.


        Returns
        -------
        bool
            Whether or not sending the post was successful.

        """
        
        self.init()
        Settings.print("post > {}".format(self.get_text()))
        if not self.get_files() and self.get_text() == "":
            Settings.err_print("Missing files and text!")
            return False
        return Driver.post(self.get_post())
            



from datetime import datetime

from ..util import defaults as DEFAULT
from ..util.settings import Settings

class Schedule:

    def __init__(self):
        self._initialized_ = False
        self.date = None
        self.time = None
        ##
        self.hour = "00"
        self.minute = "00"
        self.year = "0"
        self.month = "0"
        self.day = "0"
        self.suffix = "am"
        ##
        self.init()

    def init(self):
        """Initialize the schedule's settings"""

        if self._initialized_: return
        Settings.dev_print("initiliazing schedule...")
        schedule = Settings.get_schedule()
        date = datetime.strptime(str(schedule), DEFAULT.SCHEDULE_FORMAT)
        self.year = date.year
        self.month = date.strftime("%B")
        self.day = date.day
        self.hour = date.hour
        self.minute = date.minute
        self.suffix = "am"
        if int(self.hour) > 12:
            self.suffix = "pm"
            self.hour = int(self.hour) - 12
        Settings.dev_print("year: {}".format(self.year))
        Settings.dev_print("month: {}".format(self.month))
        Settings.dev_print("day: {}".format(self.day))
        Settings.dev_print("hour: {}".format(self.hour))
        Settings.dev_print("minutes: {}".format(self.minute))
        Settings.dev_print("suffix: {}".format(self.suffix))
        Settings.dev_print("initiliazed schedule")
        self._initialized_ = True

    def get(self):
        """
        Get the schedule's values in a dict.

        Returns
        -------
        dict
            A dict containing the values of the schedule

        """

        return dict({
            "date": self.get_date(),
            "time": self.get_time(),
            "hour" : self.hour,
            "minute" : self.minute,
            "year" : self.year,
            "month" : self.month,
            "day" : self.day,
            "suffix" : self.suffix
        })

    def get_date(self):
        """
        Gets the date value if not none else sets it from args or prompts.

        Returns
        -------
        str
            The date as a valid date string

        """

        if self.date: return self.date
        self.date = Settings.get_date()
        return self.date

        # prompt skip
        # confirm date
        if not Settings.confirm(date): return self.get_date()
        self.date = date
        return self.date

    def get_time(self):
        """
        Gets the time value if not none else sets it from args or prompts.

        Returns
        -------
        str
            The time as a valid time string

        """

        if self.time: return self.time
        # retrieve from args and return if exists
        self.time = Settings.get_time()
        return self.time



        # # if time: 
        # time = datetime.strptime(str(time), DEFAULT.SCHEDULE_FORMAT)
        # # Settings.dev_print(time)
        # time = time.strftime("%I:%M %p")
        # # Settings.dev_print(time)
        # self.time = time
        # return self.time

        # retrieve time from schedule args and return if exists
        schedule = Settings.get_schedule() or None
        if schedule:
            time = datetime.strptime(str(schedule), DEFAULT.SCHEDULE_FORMAT)
            # Settings.dev_print(time)
            time = time.strftime("%I:%M %p")
            # Settings.dev_print(time)
            self.time = time
            return self.time
        # prompt skip
        # confirm time
        if not Settings.confirm(time): return self.get_time()
        self.time = time
        return self.time

    def validate(self):
        """
        Determines whether or not the schedule settings are valid.

        Returns
        -------
        bool
            Whether or not the schedule is valid

        """

        Settings.dev_print("validating schedule...")
        today = datetime.strptime(str(datetime.now().strftime(DEFAULT.SCHEDULE_FORMAT)), DEFAULT.SCHEDULE_FORMAT)
        # schedule = datetime.strptime(str(Settings.get_schedule().now().strftime(DEFAULT.SCHEDULE_FORMAT)), DEFAULT.SCHEDULE_FORMAT)
        schedule = Settings.get_schedule()
        if not schedule: return False
        if isinstance(schedule, str):
            schedule = datetime.strptime(schedule, DEFAULT.SCHEDULE_FORMAT)
        # should invalidate if all default settings
        if str(self.get_date()) == DEFAULT.DATE and (str(self.get_time()) == DEFAULT.TIME or str(self.get_time()) == DEFAULT.TIME_NONE):
            Settings.dev_print("invalid schedule! (default date and time)")
            return False
        # cannot post in the past
        # TODO: possibly add margin of error if necessary
        elif schedule <= today:
            Settings.dev_print("invalid schedule! (must be in future)")
            return False
        Settings.dev_print("valid schedule!")
        return True



##
# unused, from Driver
##
# LOGIN_FORM = "b-loginreg__form"
# TWITTER_LOGIN0 = "//a[@class='g-btn m-rounded m-flex m-lg']"
# TWITTER_LOGIN1 = "//a[@class='g-btn m-rounded m-flex m-lg btn-twitter']"
# TWITTER_LOGIN2 = "//a[@class='btn btn-default btn-block btn-lg btn-twitter']"
# TWITTER_LOGIN3 = "//a[@class='g-btn m-rounded m-flex m-lg m-with-icon']"
# USERNAME_XPATH = "//input[@id='username_or_email']"
# PASSWORD_XPATH = "//input[@id='password']"
# SEND_BUTTON_XPATH = "//button[@type='submit' and @class='g-btn m-rounded']"
# SEND_BUTTON_CLASS2 = "button.g-btn.m-rounded"
# LIVE_BUTTON_CLASS = "b-make-post__streaming-link"
# DISCOUNT_INPUT = "form-control.b-fans__trial__select-wrapper"
# ONLYFANS_PRICE2 = "button.b-chat__btn-set-price"

ELEMENTS = [

    ### login
    {
        "name": "login",
        "classes": [],
        "text": [],
        "id": []
    },

    # username
    {
        "name": "loginUsername",
        "classes": [],
        "text": [],
        "id": []
    },

    # password
    {
        "name": "loginPassword",
        "classes": [],
        "text": [],
        "id": []
    },

    {
        "name": "loginCheck",
        "classes": ["b-make-post__streaming-link"],
        "text": [],
        "id": []
    },

    {
        "name": "rememberMe",
        "xpath": ["//input[@id='remember']"],
        "classes": [],
        "text": [],
        "id": []
    }
]


import argparse, os
import validators
from datetime import datetime
from . import defaults as DEFAULT

# Validators

#
# Args

# def valid_action(s):
# 	try:
# 		if str(s) in DEFAULT.ACTIONS:
# 			return str(s)
# 	except ValueError:
# 		msg = "Not a valid action: '{0}'.".format(s)
# 		raise argparse.ArgumentTypeError(msg)

def valid_amount(s):
	try:
		if str(s) == "max": return DEFAULT.DISCOUNT_MAX_AMOUNT
		elif str(s) == "min": return DEFAULT.DISCOUNT_MIN_AMOUNT
		elif int(s) >= DEFAULT.DISCOUNT_MIN_AMOUNT and int(s) <= DEFAULT.DISCOUNT_MAX_AMOUNT:
			return int(s)
	except ValueError:
		msg = "Not a valid discount amount: '{0}'.".format(s)
		raise argparse.ArgumentTypeError(msg)

def valid_date(s):
	try: return datetime.strptime(s, DEFAULT.DATE_FORMAT)
	except ValueError:
		msg = "Not a valid date: '{0}'.".format(s)
		raise argparse.ArgumentTypeError(msg)

def valid_duration(s):
	try:
		if str(s) == "max": return DEFAULT.DURATION_ALLOWED[-1]
		elif str(s) == "min": return DEFAULT.DURATION_ALLOWED[0]
		elif str(s) in DEFAULT.DURATION_ALLOWED: return str(s)
	except ValueError:
		msg = "Not a valid duration: '{0}'.".format(s)
		raise argparse.ArgumentTypeError(msg)
	return int(s)

def valid_promo_duration(s):
	try:
		if str(s) in DEFAULT.PROMOTION_DEFAULT.DURATION_ALLOWED: return str(s)
	except ValueError:
		msg = "Not a valid promo duration: '{0}'.".format(s)
		raise argparse.ArgumentTypeError(msg)
	return int(s)

def valid_promo_expiration(s):
	try:
		if int(s) in DEFAULT.PROMOTION_EXPIRATION_ALLOWED: return int(s)
	except ValueError:
		msg = "Not a valid promo expiration: '{0}'.".format(s)
		raise argparse.ArgumentTypeError(msg)

def valid_expiration(s):
	try:
		if str(s) == "max": return DEFAULT.EXPIRATION_MAX
		elif str(s) == "min": return DEFAULT.EXPIRATION_MIN
		elif int(s) <= DEFAULT.EXPIRATION_MAX: return int(s)
	except ValueError:
		msg = "Not a valid expiration: '{0}'.".format(s)
		raise argparse.ArgumentTypeError(msg)

def valid_limit(s):
	try:
		if str(s) == "max": return DEFAULT.LIMIT_ALLOWED[-1]
		elif str(s) == "min": return DEFAULT.LIMIT_ALLOWED[0]
		elif int(s) in DEFAULT.LIMIT_ALLOWED: return int(s)
	except ValueError:
		msg = "Not a valid limit: '{0}'.".format(s)
		raise argparse.ArgumentTypeError(msg)
	return int(s)

def valid_month(s):
	try:
		if str(s) == "max": return DEFAULT.DISCOUNT_MAX_MONTHS
		elif str(s) == "min": return DEFAULT.DISCOUNT_MIN_MONTHS
		elif int(s) >= DEFAULT.DISCOUNT_MIN_MONTHS and int(s) <= DEFAULT.DISCOUNT_MAX_MONTHS:
			return int(s)
	except ValueError:
		msg = "Not a valid month number: '{0}'.".format(s)
		raise argparse.ArgumentTypeError(msg)

def valid_path(s):
	try:
		if isinstance(s, list):
			for f in s: os.stat(s)
		else: os.stat(s)
	except FileNotFoundError:
		# check as url
		return valid_url(s)
		# msg = "Not a valid path: '{0}'.".format(s)
		# raise argparse.ArgumentTypeError(msg)
	return s

def valid_url(s):
	if not validators.url(s):
		msg = "Not a valid path or url: '{0}'.".format(s)
		raise argparse.ArgumentTypeError(msg)
	return s

def valid_price(s):
	if str(s) == "max": return DEFAULT.PRICE_MAXIMUM
	elif str(s) == "min": return DEFAULT.PRICE_MINIMUM
	try: return "{:.2f}".format(float(s))
	except ValueError:
		msg = "Not a valid price: '{0}'.".format(s)
		raise argparse.ArgumentTypeError(msg)

def valid_schedule(s):
	try: return datetime.strptime(s, DEFAULT.SCHEDULE_FORMAT)
	except ValueError:
		msg = "Not a valid schedule: '{0}'.".format(s)
		raise argparse.ArgumentTypeError(msg)

def valid_time(s):
	try: return datetime.strptime(s, DEFAULT.TIME_FORMAT)
	except ValueError:
		msg = "Not a valid time: '{0}'.".format(s)
		raise argparse.ArgumentTypeError(msg)

# check against min/max amounts & months
# def valid_discount(s):
  # pass

# def valid_category(s):
#   if str(s) not in DEFAULT.CATEGORIES_DEFAULT:
#     msg = "Not a valid category: '{0}'.".format(s)
#     raise argparse.ArgumentTypeError(msg)



import argparse
from typing import Dict, Any

args: Dict[str, Any] = {}

########################################################################################################

##
# Argument Parser
##

parser = argparse.ArgumentParser(prog='snarf', allow_abbrev=False, epilog="Shnarrf!", 
  description="No mention of old Shnarf, I notice. Go ahead, just take all the glory, and leave it to Snarf to clean up after you. I don't mind!", conflict_handler='resolve')

############

import os
if os.environ.get("ENV") != "test":
  from .optional_args import apply_subcommand_args
  apply_subcommand_args(parser)
elif os.environ.get("ENV") == "test":
  from .optional_args import apply_shim_args
  apply_shim_args(parser)

from .optional_args import apply_args
apply_args(parser)

##
import pkg_resources
parser.version = str(pkg_resources.get_distribution("onlysnarf").version)
parser.add_argument('-version', action='version')

############################################################################################

try:
  parsedargs, unknownargs = parser.parse_known_args()
  # print("unknown args: {}".format(unknownargs))
  args.update(vars(parsedargs))
except Exception as e:
    print(e)
    print("Error: Incorrect arg format")
    parser.exit(1)

#############
# Debugging #
# import sys
# print(args)
# sys.exit(0)